local L = RC_locale()
local addonName = "RetardCheck"
local LaddonName = L[addonName]
RetardCheck = LibStub("AceAddon-3.0"):NewAddon(addonName)
local addon = RetardCheck 

local defaults = {
  profile = {
    general = {
      debug = false, -- for addon debugging
    },
    minimap = {
        hide = false,
    },
    checks = {
      empty = true,
      fishingpole = true,
      ilvl = true,
      durability = true,
      --beltbuckle = true,
      enchant = true,
      profenchant = true,
      talentpts = true,
      spellhit = true,
      meleehit = true,
      rangedhit = true,
      expertise = true,
    },
    thresh = {
      ilvl = 250,
      durability = 0.05,
      spellhit = 0,
      meleehit = 0,
      rangedhit = 0,
      expertise = 0,
    },
  }
}

local settings = defaults.profile
local optionsFrame
local charName
local hiddenFrame = CreateFrame("Button", addonName.."HiddenFrame", UIParent)
local revision = tonumber(("$Revision: 3 $"):match("%d+"))
local minimapIcon = LibStub("LibDBIcon-1.0")
local LDB, LDBo

local function chatMsg(msg)
     DEFAULT_CHAT_FRAME:AddMessage(addonName..": "..msg)
end
local function debug(msg)
  if addon.db.profile.general.debug then
     chatMsg(msg)
  end
end

function addon:myOptions() 
return {
  type = "group",
  set = function(info,val)
          local s = settings ; for i = 1,#info-1 do s = s[info[i]] end
          s[info[#info]] = val; debug(info[#info].." set to: "..tostring(val))
          addon:Update()
        end,
  get = function(info)
          local s = settings ; for i = 1,#info-1 do s = s[info[i]] end
          return s[info[#info]] end,
  args = {
   general = {
    type = "group",
    inline = true,
    name = L["General"],
    args = {
      debug = {
        name = L["Debug"],
        desc = L["Toggle debugging output"],
        type = "toggle",
        guiHidden = true,
      },
      config = {
        name = L["Config"],
        desc = L["Open the configuration GUI"],
        type = "execute",
        guiHidden = true,
        func = function() addon:Config() end,
      },
      minimap = {
        order = 15,
        name = L["Minimap Icon"],
        desc = L["Display minimap icon"],
        type = "toggle",
        set = function(info,val)
          settings.minimap.hide = not val
          addon:Update()
	end,
        get = function() return not settings.minimap.hide end,
      },
    },
   },
   checks = {
	name = L["Checks"],
	type = "group",
	args = (function()
	  local ret = {}
	  for o,v in pairs(addon.checks_ordered) do
	    ret[v.checkname] = {
	      order = o,
	      name = v.failname,
	      desc = v.desc or v.failname,
	      type = "toggle",
	    }
	    if v.threshdesc then
	      ret[v.checkname].order = o*10 + 10002
	      ret[v.checkname.."_header"] = {
	        order = o*10 + 10001,
		type = "header",
		name = "",
		cmdHidden = true,
	      }
	      ret[v.checkname.."_thresh"] = {
	        order = o*10 + 10003,
		name = v.threshname or v.failname.." "..L["threshold"],
		desc = v.threshdesc,
		step = v.threshstep,
		isPercent = v.threshpercent,
		disabled = function() return not settings.checks[v.checkname] end,
                type = "range",
		min = v.threshmin,
		max = v.threshmax,
                get = function() return settings.thresh[v.checkname] end,
                set = function(info,val)
		    settings.thresh[v.checkname] = val
                    addon:Update()
	        end,
	      }
	    end
	  end
	  return ret
	end)(),
     },
  }
} 
end

local function table_clone(t)
  if not t then return nil
  elseif type(t) == "table" then
    local res = {}
    for k,v in pairs(t) do
      res[table_clone(k)] = table_clone(v)
    end
    return res
  else
    return t
  end
end

local function order_table(t) -- sort and compact a table by order field
    local tmp = {}
    local index = {}
    for k,v in pairs(t) do
       v.checkname = k 
       if v.order then
         if tmp[v.order] then
           debug("ERROR: duplicate table order value: "..v.order)
         else
           tmp[v.order] = v
         end
	 table.insert(index, v.order)
       end
    end
    local max = table.maxn(tmp)
    for k,v in pairs(t) do
       if not v.order then
          max = max + 1
          table.insert(retval, max, v)
          v.order = max
	  table.insert(index, v.order)
       end
    end
    table.sort(index)
    local retval = {}
    for i,o in ipairs(index) do
      retval[i] = tmp[o]
      retval[i].order = i
    end
    return retval
end

-- -------------------------------------------------------------------------------------------------
-- Item queries and helper functions
local FPsubclass
function addon:isFishingPole(itemid)
  if not FPsubclass then
    FPsubclass = select(7, GetItemInfo(6256))
  end
  return select(7, GetItemInfo(itemid)) == FPsubclass
end

local OHsubclass
function addon:isHeldInOffhand(itemid)
  if not OHsubclass then
    OHsubclass = select(7, GetItemInfo(43656))
  end
  return select(7, GetItemInfo(itemid)) == OHsubclass
end
   
local TwoHandsubclass = nil
function addon:is2HWeapon(itemid)
  if not itemid then return false end
  if not TwoHandsubclass then
      TwoHandsubclass = {}
      for _,id in pairs({2491, 2489, 2495, 2493, 6367, 44654}) do
        local subclass = select(7, GetItemInfo(id))
        if subclass and TwoHandsubclass then -- might fail on first call due to lag
          TwoHandsubclass[subclass] = true
        else
          TwoHandsubclass = nil
        end
      end
  end
  local subclass = select(7, GetItemInfo(itemid))
  if TwoHandsubclass then
     return TwoHandsubclass[subclass]
  else
     return true -- conservative don't know
  end
end

local PID_BS   = 2018   -- XXX check for 2 slots
local PID_JC   = 25229  -- XXX count JC gems
local PID_ENG  = 4036   -- XXX figure out eng shit
local PID_ENCH = 7411
local PID_INSC = 45357
local PID_LW   = 2108
local PID_TAIL = 3908

local profenchants = {
  [4115]=1, [4116]=1, [4118]=1, -- Tailoring
  [4078]=1, [4079]=1, [4080]=1, [4081]=1, -- Enchanting
  [4189]=1, [4190]=1, [4191]=1, [4192]=1, -- Leatherworking
  [4193]=1, [4194]=1, [4195]=1, [4196]=1, -- Inscription
  [4179]=1, [4180]=1, [4181]=1, [4182]=1, [4183]=1, -- Engineering Gloves
  [4187]=1, [4188]=1, -- Engineering Belt
}

-- returns rank
function addon:hasProfession(pid)
  local pname = GetSpellInfo(pid)
  for _,profid in pairs({ GetProfessions() }) do
    local pn, _, rank, rankmax = GetProfessionInfo(profid)
    if (pn == pname) then return rank end
  end
  return 0
end

-- Color, Ltype, Id, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, reforging, Name
function addon:decomposeItemLink(itemlink)
  return string.match(itemlink, "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")
end  

-- returns raid role for the unit (TANK/HEALER/DAMAGER)
function addon:raidRole(unitid)
  local currrole = UnitGroupRolesAssigned(unitid)
  if currrole and currrole ~= "NONE" then
    return currrole
  end
  if GetPartyAssignment("MAINTANK", unitid) then
    return "TANK"
  end
  local inspect = not UnitIsUnit(unitid,"player")
  local tabIndex = GetPrimaryTalentTree(inspect, false)
  if not tabIndex then return end -- untalented
  local role1,role2 = GetTalentTreeRoles(tabIndex,inspect,false)
  if not role2 then
    return role1
  else -- more than one possibility (eg feral druid)
    local _, class = UnitClass(unitid)
    if class == "DRUID" then
      local tanktalents = 0 -- look for tank talents
      for ti = 1, GetNumTalents(tabIndex, inspect, nil) do
        local name, _, _, _, rank, maxrank = GetTalentInfo(tabIndex, ti, inspect, nil, nil)
        if (name == "Thick Hide" or name == "Natural Reaction") and rank == maxrank then
          tanktalents = tanktalents + 1
        end
      end
      if tanktalents >= 2 then
        return "TANK"
      else
        return "DAMAGER"
      end
    end
    return nil
  end
end

local dpsstats = {
    MAGE     = { spellhit=true },
    WARLOCK  = { spellhit=true },
    PRIEST   = { spellhit=true },
    HUNTER   = { rangedhit=true },
    DEATHKNIGHT = { meleehit=true, spellhit=true, expertise=true },
    ROGUE    = { meleehit=true, spellhit=true, expertise=true },
    PALADIN  = { meleehit=true, expertise=true },
    WARRIOR  = { meleehit=true, expertise=true },
    SHAMAN   = { [1] = { spellhit=true }, [2] = { meleehit=true, spellhit=true, expertise=true } }, 
    DRUID    = { [1] = { spellhit=true }, [2] = { meleehit=true, expertise=true } }, 
}

function addon:dpsStat(unitid, statname)
    local _, class = UnitClass(unitid)
    local inspect = not UnitIsUnit(unitid,"player")
    local tabIndex = GetPrimaryTalentTree(inspect, false)
    local stats = dpsstats[class][tabIndex] or dpsstats[class]
    return stats[statname]
end

function addon:InitData()
  -- init
  if not addon.itemchecks_ordered then -- first call
    for name,data in pairs(addon.itemslots) do
      data.id = GetInventorySlotInfo(name)
      data.shortname = name:gsub("Slot",""):gsub("[01]",""):gsub("Secondary","Off")
    end
    addon:InitChecks()
    addon.checks_ordered = order_table(addon.checks)
  end

  addon:isFishingPole(6256)   
  addon:is2HWeapon(6256)   
  addon:isHeldInOffhand(6256) 
end
-- -------------------------------------------------------------------------------------------------
-- slot table
addon.itemslots = {
	ChestSlot    = { enchantable=true }, -- INVSLOT_CHEST
	FeetSlot     = { enchantable=true }, -- INVSLOT_FEET
	WristSlot    = { enchantable=true }, -- INVSLOT_WRIST
	HandsSlot    = { enchantable=true }, -- INVSLOT_HAND
	HeadSlot     = { enchantable=true }, -- INVSLOT_HEAD
	LegsSlot     = { enchantable=true }, -- INVSLOT_LEGS
	ShoulderSlot = { enchantable=true }, -- INVSLOT_SHOULDER
	BackSlot     = { enchantable=true }, -- INVSLOT_BACK
	MainHandSlot = { enchantable=true }, -- INVSLOT_MAINHAND
	SecondaryHandSlot = { enchantable=true }, -- INVSLOT_OFFHAND
	RangedSlot   = { enchantable=true }, -- INVSLOT_RANGED
	Finger0Slot  = {  }, -- INVSLOT_FINGER1
	Finger1Slot  = {  }, -- INVSLOT_FINGER2
	WaistSlot    = {  }, -- INVSLOT_WAIST
	NeckSlot     = {  }, -- INVSLOT_NECK
	Trinket0Slot = {  }, -- INVSLOT_TRINKET1
	Trinket1Slot = {  }, -- INVSLOT_TRINKET2
}
-- -------------------------------------------------------------------------------------------------
-- Item Checks
local function statCheck(Order, Statname, Failchance, Failname, Threshdesc)
 return {
    order = Order,
    failname = Failname,
    playeronly = true,
    threshdesc = Threshdesc,
    threshmin = 0,
    threshmax = 0.20,
    threshstep = 0.005,
    threshpercent = true,
    threshname = L["Miss allowance"],
    playercheck = function(unitid) 
      if addon:raidRole(unitid) ~= "DAMAGER" or -- not a dps
         not addon:dpsStat(unitid, Statname) then -- not a relevant stat
	 return true
      end
      local max = 100*settings.thresh[Statname]
      local lvldelta
      if UnitInRaid(unitid) then
        lvldelta = 3
      else
        lvldelta = 2
      end
      local a,b = Failchance(lvldelta)
      if b and b > a then a = b end -- offhand dodge
      if a > max then
        return false, string.format("%.02f%% "..MISS, a)
      else
        return true
      end
    end,
  }
end

function addon:InitChecks() 
addon.checks = {

  empty = {
    order = 10,
    failname = L["Empty Slot"],
    itemstop = true, -- stop checking this slot on failure
    itemcheck = function(itemid, itemlink, slotid)
      return itemid and itemlink
    end,
  },

  fishingpole = {
    order = 20,
    failname = L["Fishing Pole"],
    slots = { MainHand },
    itemstop = true, -- stop checking this slot on failure
    itemcheck = function(itemid, itemlink, slotid)
        return not addon:isFishingPole(itemid) 
    end,
  },

  ilvl = {
    order = 30,
    failname = L["Item Ilvl"],
    threshdesc = L["Minimum equipped item level to allow"],
    threshmin = 25,
    threshmax = 400,
    threshstep = 1,
    itemstop = true, -- stop checking this slot on failure
    itemcheck = function(itemid, itemlink, slotid)
      local min = settings.thresh.ilvl
      local quality, ilvl = select(3, GetItemInfo(itemid))
      return (ilvl and ilvl >= min) or (quality == 7) -- allow heirlooms
      --- XXX: tooltip scan for ITEM_LEVEL_RANGE_CURRENT to determine level appropriate
    end,
  },

  durability = {
    order = 40,
    playeronly = true,
    failname = L["Item Durability"],
    threshdesc = L["Minimum item durability percentage to allow"],
    threshmin = 0,
    threshmax = 1,
    threshpercent = true,
    itemcheck = function(itemid, itemlink, slotid)
      local cur, max = GetInventoryItemDurability(slotid)
      if not cur then return true end -- slot with no repair or info not available
      local val = cur/max
      local min = settings.thresh.durability
      return val >= min
    end,
  },

  enchant = {
    order = 50,
    failname = L["Item Enchant"],
    slots = (function() local t = {} 
             for _,s in pairs(addon.itemslots) do
                if s.enchantable then
		  table.insert(t, s.shortname)
		end
	     end
             return t
	    end)(),
    itemcheck = function(itemid, itemlink, slotid, unitid)
      local enchant = select(4,addon:decomposeItemLink(itemlink))
      local _, class = UnitClass(unitid)
      if enchant and tonumber(enchant) and tonumber(enchant) > 0 then
        return true -- XXX: check for "good" enchants
      elseif slotid == INVSLOT_OFFHAND then
        return not addon:isHeldInOffhand(itemid)
      elseif slotid == INVSLOT_RANGED and class ~= "HUNTER" then
        return true 
      else
        return false
      end
    end,
  },

  profenchant = {
    order = 60,
    playeronly = true,
    failname = L["Profession Enchant"],
    slots = { Finger0, Finger1, Wrist, Shoulder, Back },
    itemcheck = function(itemid, itemlink, slotid)
      local enchant = select(4,addon:decomposeItemLink(itemlink))
      enchant = (enchant and tonumber(enchant)) or 0
      if slotid == INVSLOT_FINGER1 or slotid == INVSLOT_FINGER2 then
        if addon:hasProfession(PID_ENCH) >= 475 and not profenchants[enchant] then
	  return false
	end
      elseif slotid == INVSLOT_BACK then
        if addon:hasProfession(PID_TAIL) >= 500 and not profenchants[enchant] then
	  return false
	end
      elseif slotid == INVSLOT_WRIST then
        if addon:hasProfession(PID_LW) >= 500 and not profenchants[enchant] then
	  return false
	end
      elseif slotid == INVSLOT_SHOULDER then
        if addon:hasProfession(PID_INSC) >= 500 and not profenchants[enchant] then
	  return false
	end
      end
      return true
    end,
  },

  --[[
  beltbuckle = { -- XXX need to scan bare itemlink and count sockets
    order = ,
    failname = L["Belt Buckle"],
    itemstop = true, -- stop checking this slot on failure
    slots = { Waist },
    itemcheck = function(itemid, itemlink, slotid)
      local buck = select(4,addon:decomposeItemLink(itemlink))
      return (buck == 417)
    end,
  },
  ]]--

  talentpts = {
    order = 1000,
    failname = L["Talent points"],
    playercheck = function(unitid) 
      local unspent = GetUnspentTalentPoints(not UnitIsUnit(unitid,"player"),false, nil)
      return (not unspent or unspent == 0), unspent and unspent.." "..L["unspent talent points"]
    end,
  },

  spellhit =  statCheck(1100, "spellhit", GetSpellMissChance,  L["Spell Hit"], L["Maximum spell miss chance to allow"]),
  meleehit =  statCheck(1110, "meleehit", GetMeleeMissChance,  L["Melee Hit"], L["Maximum melee miss chance to allow"]),
  rangedhit = statCheck(1120, "rangedhit", GetRangedMissChance, L["Ranged Hit"], L["Maximum ranged miss chance to allow"]),
  expertise = statCheck(1130, "expertise", GetEnemyDodgeChance, L["Expertise"], L["Maximum dodge chance to allow"]),
}
end
-- -------------------------------------------------------------------------------------------------
function addon:RunCheck()
  debug("RunCheck")
  addon:ResetFails("player")
  addon:CheckUnit("player")
  addon:ReportFails("player")
end
function addon:CheckUnit(unitid)
  local isplayer = UnitIsUnit(unitid, "player")
  -- item checks
  for name,data in pairs(addon.itemslots) do
    local itemid = GetInventoryItemID(unitid, data.id)
    local itemlink = GetInventoryItemLink(unitid, data.id)

    if not itemid and data.id == INVSLOT_OFFHAND and 
       addon:is2HWeapon(GetInventoryItemID(unitid, INVSLOT_MAINHAND)) --- XXX: titan's grip
    then
      -- offhand is empty for two-hander, skip checks
    else
     for _,check in ipairs(addon.checks_ordered) do
       local doit = settings.checks[check.checkname]
       if doit and check.slots then
         doit = false
         for _,s in pairs(check.slots) do
	   if s == name or s == data.shortname then
	     doit = true
	     break
	   end
	 end
       end
       if doit and check.playeronly and not isplayer then
         doit = false
       end
       if doit and not itemid and check.checkname ~= "empty" then
         doit = false
       end
       if doit and check.itemcheck then
         --debug("Checking "..data.shortname.." "..itemid.." "..itemlink)
         local passed = check.itemcheck(itemid, itemlink, data.id, unitid)
         if not passed then
	   addon:RecordFail(unitid, check, name, itemlink or "")
           if check.itemstop then
	     break
	   end
	 end
       end
     end
    end
  end
  -- player checks
  for _,check in ipairs(addon.checks_ordered) do
     local doit = settings.checks[check.checkname]
     if doit and check.playeronly and not isplayer then
       doit = false
     end
     if doit and check.playercheck then
       local passed, note = check.playercheck(unitid)
       if not passed then
	 addon:RecordFail(unitid, check, nil, nil, note)
       end
     end
  end
end

addon.failtable = {}
function addon:ResetFails(unitid)
  addon.failtable[unitid] = addon.failtable[unitid] or {}
  wipe(addon.failtable[unitid])
end

function addon:RecordFail(unitid, check, slot, itemlink, note)
  addon.failtable[unitid] = addon.failtable[unitid] or {}
  local faildata = addon.failtable[unitid][check.order] or {}
  faildata.check = check
  faildata.note = note
  if slot then
    faildata.slot = faildata.slot or {}
    faildata.slot[slot] = itemlink
  end
  addon.failtable[unitid][check.order] = faildata
  addon.failtable[unitid].count = (addon.failtable[unitid].count or 0) + 1
end

function addon:ReportFails(unitid)
  local ft = addon.failtable[unitid]
  if ft.count then
    print(addonName.." "..string.format(L["Detected %i failures for %s"],ft.count,UnitName(unitid))..":")
    for _, check in ipairs(addon.checks_ordered) do
       local faildata = ft[check.order]
       if faildata then
         local str = "   \124cffff0000"..faildata.check.failname.."\124r"
         if faildata.slot then
	   local cnt = 0
	   str = str..": "
	   for slot,itemlink in pairs(faildata.slot) do
	     cnt = cnt + 1
	     if cnt > 1 then str = str..", " end
	     str = str..itemlink.." ("..addon.itemslots[slot].shortname..")"
	     if cnt > 2 then print(str); str = "      "; cnt = 0 end
           end
	   if #str > 10 then
	     print(str)
	   end
         else
	   if faildata.note then
	     str = str.." "..faildata.note
	   end
           print(str)
         end
       end
    end
    return ft.count
  else 
    debug("No failures for "..UnitName(unitid))
    return 0
  end
end
-- -------------------------------------------------------------------------------------------------

function addon:RefreshConfig()
  -- things to do after load or settings are reset
  debug("RefreshConfig")
  settings = addon.db.profile
  addon.settings = settings
  charName = UnitName("player")
  for k,v in pairs(defaults.profile) do
     if settings[k] == nil then
       settings[k] = table_clone(v)
     end
  end
  settings.loaded = true
  addon:Update()
end

function addon:Update()
  -- things to do when settings change
  if LDBo then
    if settings.minimap.hide then
      minimapIcon:Hide(addonName)
    else
      minimapIcon:Show(addonName)
    end
  end
end

function addon:OnInitialize()
  addon.db = LibStub("AceDB-3.0"):New("RetardCheckDB", defaults)
  addon:RefreshConfig()
  addon:InitData()

  local options = addon:myOptions()
  LibStub("AceEvent-3.0"):Embed(addon)
  LibStub("AceConfigRegistry-3.0"):ValidateOptionsTable(options, addonName)
  LibStub("AceConfig-3.0"):RegisterOptionsTable(addonName, options, {"retardcheck", "rc"})
  optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName, addonName, nil, "general")
  optionsFrame.default = function()
       for k,v in pairs(defaults.profile) do settings[k] = table_clone(v) end
       addon:RefreshConfig()
       if InterfaceOptionsFrame:IsShown() then
         addon:Config(); addon:Config()
       end
  end
  local checksFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName, L["Checks"], addonName, "checks")
  checksFrame.default = optionsFrame.default
  options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(addon.db)
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName, L["Profiles"], addonName, "profiles")

  debug("OnInitialize")

  self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
  self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
  self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")
  self.db.RegisterCallback(self, "OnDatabaseReset", "RefreshConfig")

end

function addon:Config()
  if optionsFrame then
    if ( InterfaceOptionsFrame:IsShown() ) then
      InterfaceOptionsFrame:Hide();
    else
      InterfaceOptionsFrame_OpenToCategory(optionsFrame)
    end
  end
end

function addon:OnEnable()
  debug("OnEnable")
  addon:RegisterEvent("READY_CHECK")
 
  if LDB then
    return
  end
  if AceLibrary and AceLibrary:HasInstance("LibDataBroker-1.1") then
    LDB = AceLibrary("LibDataBroker-1.1")
  elseif LibStub then
    LDB = LibStub:GetLibrary("LibDataBroker-1.1",true)
  end
  if LDB then
    LDBo = LDB:NewDataObject(addonName, {
        type = "launcher",
        label = addonName,
        icon = "Interface\\Icons\\spell_shadow_mindtwisting",
        OnClick = function(self, button)
                if button == "RightButton" then
                        addon:Config()
                else
                        addon:RunCheck()
                end
        end,
        OnTooltipShow = function(tooltip)
                if tooltip and tooltip.AddLine then
                        tooltip:SetText(addonName)
                        tooltip:AddLine("|cffff8040"..L["Left Click"].."|r "..L["to run a check"])
                        tooltip:AddLine("|cffff8040"..L["Right Click"].."|r "..L["for options"])
                        tooltip:Show()
                end
        end,
     })
  end 

  if LDBo then
    minimapIcon:Register(addonName, LDBo, settings.minimap)
  end
  addon:Update()
end

function addon:READY_CHECK(sender)
  debug("READY_CHECK")
  addon:RunCheck()
end

